#include <ostream>
#include <math.h>
#include <assert.h>
#include "TNR_VECTOR.hxx"


//////////////////////////////////
//      TNR_VECTOR class      //
//////////////////////////////////
TNR_VECTOR::TNR_VECTOR() {
   _v[0] = _v[1] = _v[2] = 0.0;
}

TNR_VECTOR::~TNR_VECTOR() { 
   _v[0] = _v[1] = _v[2] = 0.0;
}

TNR_VECTOR::TNR_VECTOR(const double x, const double y, const double z) {
   _v[0] = x; _v[1] = y; _v[2] = z;
}

TNR_VECTOR& TNR_VECTOR::operator= (const TNR_VECTOR& p) { 
   _v[0] = p._v[0]; _v[1] = p._v[1]; _v[2] = p._v[2];
   return (*this);
}

TNR_VECTOR& TNR_VECTOR::operator+= (const TNR_VECTOR& p) {
   _v[0] += p._v[0]; _v[1] += p._v[1]; _v[2] += p._v[2];
   return (*this);
}

TNR_VECTOR& TNR_VECTOR::operator-= (const TNR_VECTOR& p) {
   _v[0] -= p._v[0]; _v[1] -= p._v[1]; _v[2] -= p._v[2];
   return (*this);
}


TNR_VECTOR& TNR_VECTOR::zero() {
   _v[0] = _v[1] = _v[2] = 0.0;
   return (*this);
}

double& TNR_VECTOR::operator[](int i) { 
   assert( ( i == 0 )||( i == 1 )||( i == 2 ) );
   return _v[i]; 
}

double& TNR_VECTOR::operator()(int i) { 
   assert( ( i == 0 )||( i == 1 )||( i == 2 ) );
   return _v[i]; 
}

void TNR_VECTOR::upper(TNR_VECTOR& p) {
   (_v[0] > p.x()) ?  _v[0] : _v[0] =  p.x();
   (_v[1] > p.y()) ?  _v[1] : _v[1] =  p.y();
   (_v[2] > p.z()) ?  _v[2] : _v[2] =  p.z();
}

void TNR_VECTOR::lower(TNR_VECTOR& p) {
   (_v[0] < p.x()) ?  _v[0] : _v[0] =  p.x();
   (_v[1] < p.y()) ?  _v[1] : _v[1] =  p.y();
   (_v[2] < p.z()) ?  _v[2] : _v[2] =  p.z();
}

double TNR_VECTOR::norm() const {
   return sqrt((double)(_v[0]*_v[0] + _v[1]*_v[1] + _v[2]*_v[2]));
}

TNR_VECTOR& TNR_VECTOR::normalize() {
   double len = norm();
   _v[0] /= len; _v[1] /= len; _v[2] /= len;
   return *this;
} 

void  TNR_VECTOR::move(const TNR_VECTOR& center, TNR_VECTOR& vector)
{
  TNR_VECTOR v = *this - center;
  v = v + vector + center;
  _v[0] = v[0];
  _v[1] = v[1];
  _v[2] = v[2];
}
void  TNR_VECTOR::scale(const TNR_VECTOR& center, TNR_VECTOR& factor)
{
  TNR_VECTOR v = *this - center;
  v = v * factor+ center;
  _v[0] = v[0];
  _v[1] = v[1];
  _v[2] = v[2];
}

TNR_VECTOR  TNR_VECTOR
::rotateByXAxis (const TNR_VECTOR& center, double angle)
{
  TNR_VECTOR v = *this - center;
  double s = sin (angle);
  double c = cos (angle);
  v = TNR_VECTOR (v.x(), c * v.y() - s * v.z(), s * v.y() + c * v.z());
  return center + v;

}

TNR_VECTOR  TNR_VECTOR
::rotateByYAxis (const TNR_VECTOR& center, double angle)
{
  TNR_VECTOR v = *this - center;
  double s = sin (angle);
  double c = cos (angle);
  v = TNR_VECTOR (c * v.x() + s * v.z(), v.y(), - s * v.x() + c * v.z());
  return center + v;
}

TNR_VECTOR  TNR_VECTOR
::rotateByZAxis (const TNR_VECTOR& center, double angle)
{
  TNR_VECTOR v = *this - center;
  double s = sin (angle);
  double c = cos (angle);
  v = TNR_VECTOR (c * v.x() - s * v.y(), s * v.x() + c * v.y(), v.z());
  return center + v;
}


TNR_VECTOR operator* (const double a, const TNR_VECTOR& p) {
   return TNR_VECTOR(a * p.x(), a * p.y(), a * p.z());
} 

TNR_VECTOR operator* (const TNR_VECTOR& p0,const TNR_VECTOR& p1)
{
  return  TNR_VECTOR(p0.x() * p1.x(), p0.y() * p1.y(), p0.z() * p1.z());
}

TNR_VECTOR operator/ (const TNR_VECTOR& p,const double a) {
   return TNR_VECTOR(p.x()/a, p.y()/a, p.z()/a); 
} 

TNR_VECTOR operator/ (const TNR_VECTOR& p1,const TNR_VECTOR& p2) {
  return TNR_VECTOR(p1.x()/p2.x(), p1.y()/p2.y(), p1.z()/p2.z());
} 

TNR_VECTOR operator+ (const TNR_VECTOR& p1, const TNR_VECTOR& p2) {
   return TNR_VECTOR(p1.x() + p2.x(), p1.y() + p2.y(), p1.z() + p2.z());
} 

TNR_VECTOR operator+ (const TNR_VECTOR& p1, const double m) {
   return TNR_VECTOR(p1.x() + m, p1.y() + m, p1.z() + m);
} 


TNR_VECTOR operator- (const TNR_VECTOR& p1, const TNR_VECTOR& p2) {
   return TNR_VECTOR(p1.x() - p2.x(), p1.y() - p2.y(), p1.z() - p2.z());
} 

TNR_VECTOR operator- (const TNR_VECTOR& p1, const double m) {
   return TNR_VECTOR(p1.x() - m, p1.y() - m, p1.z() - m);
} 

int operator== (const TNR_VECTOR& a, const TNR_VECTOR& b) {
   if ( 0.0 < fabs(a.x()-b.x()) ) return 0;
   if ( 0.0 < fabs(a.y()-b.y()) ) return 0;
   if ( 0.0 < fabs(a.z()-b.z()) ) return 0;
   return 1;
}

int operator!= (const TNR_VECTOR& a, const TNR_VECTOR& b) {
  return !(a == b);
}

double dotProduct(const TNR_VECTOR& a, const TNR_VECTOR& b) {
   return a.x() * b.x() + a.y() * b.y() + a.z() * b.z();
}

TNR_VECTOR crossProduct(const TNR_VECTOR& a, const TNR_VECTOR& b) {
   return TNR_VECTOR(
       a.y() * b.z() - a.z() * b.y(), 
       a.z() * b.x() - a.x() * b.z(), 
       a.x() * b.y() - a.y() * b.x() );
}

double distanceBetween(const TNR_VECTOR& a, const TNR_VECTOR& b) {
  return (a-b).norm();
}




double angleBetween(const TNR_VECTOR& vec01, const TNR_VECTOR& vec02)
{

  double norm1 = vec01.norm(); 
  double norm2 = vec02.norm();
 
  assert(norm1 > 0.0); 
  assert(norm2 > 0.0); 

  double temp = dotProduct(vec01,vec02);

  if(temp == 0.0)  return  (0.5 * M_PI); 

#if 0
  if(TNR_range_simple_match(temp,0.0))  return  (0.5 * M_PI); 
#endif

  double norm = norm1 * norm2;

  assert(norm > 0.0); 

  double c = temp/norm;

  if(1.0 < c) c = 1.0;
  if(-1.0 > c) c = -1.0;

  return (acos(c));
}
